import traer.physics.*;
import traer.animation.*;
//import org.json.*;
/// import moviemaker.*;

final float NODE_SIZE = 10;
final float EDGE_LENGTH = 20;
final float EDGE_STRENGTH = 0.2;
final float SPACER_STRENGTH = 500;

SmartSystem textSystem, boxSystem, boxSizeSystem, floatSystem;

int frame;
PFont htmlFont;
PFont renderFont, renderFontItal;
String htmlText[] = {"<html>","<body>","<p>this is just plain text,</p>","<p>but text with <em>structure</em></p>","<p>and <a href=\"foo.html\">connections</a></p>","</body>","</html>"};
color bgColor;


class SmartSystem {
    ParticleSystem physics;
    ArrayList names; // Holds an id string for each particle
    ArrayList centroids;
    ArrayList attribs; // Holds a hashtable of attribs for each particle
    Animator anim;
    
    SmartSystem(float gravity, float drag, float smoothness){
        anim = new Animator(smoothness);
        physics = new ParticleSystem(gravity, drag);
        names = new ArrayList();
        centroids = new ArrayList();
    }

    void clear(){
        physics.clear();
        names.clear();
        centroids.clear();
    }
    
    int count(){
        return physics.numberOfParticles();
    }
    
    ParticleSystem sys(){
        return physics;
    }
    
    void dbgState(){
        int numP = physics.numberOfParticles();
        for (int i=0; i<numP; i++){
            Particle p = physics.getParticle(i);
            println("p "+names.get(i)+": "+p.position().x()+","+p.position().y());
        }
    }

    void linkNodes(String nodeA, String nodeB){
        Particle a = getParticle(nodeA);
        Particle b = getParticle(nodeB);

        for ( int i = 0; i < physics.numberOfParticles(); ++i )
        {
            Particle q = physics.getParticle( i );
            if ( a != q && a != b )
            physics.makeAttraction( a, q, -SPACER_STRENGTH, 20 );
        }
        
        physics.makeSpring( a, b, EDGE_STRENGTH, EDGE_STRENGTH, EDGE_LENGTH );
    }

    
    Particle makeParticle(String particleName, float x, float y){
        Particle newP = physics.makeParticle();
        newP.moveTo(x,y,0);        
        names.add(particleName);

        Smoother2D cent = anim.make2DSmoother();
        cent.setValue(x,y);
        cent.setTarget(-1,-1);
        centroids.add(cent);
        
        physics.tick(0.0);
        return newP;
    }
    
    Particle getParticle(String particleName){
        Particle p = null;
        int idx = names.indexOf(particleName);
        if (idx != -1){
            p = physics.getParticle(idx);
        }else{
            println("Couldn't find a particle named "+particleName);
        }
        return p;
    }
    
    Particle getParticle(int i){
        Particle p = null;
        if (i<physics.numberOfParticles()){
            p = physics.getParticle(i);
        }else{
            println("Only "+ physics.numberOfParticles() +" particles in system. "+ i +" is out of range");
        }
        return p;
    }
    
    void delParticle(String particleName){
        int idx = names.indexOf(particleName);
        delParticle(idx);
    }

    void delParticle(int i){
        if (i!=-1 && i < physics.numberOfParticles()){
            Particle p = getParticle(i);
            p.kill();
            names.remove(i);
            centroids.remove(i);
            physics.tick(0.0);        
        }else{
            println("Only "+ physics.numberOfParticles() +" particles in system. "+ i +" is out of range. Can't delete.");
        }
    }
    
    String nameForParticle(Particle mysteryP){
        for (int i=0; i<physics.numberOfParticles(); i++){
            if (mysteryP == physics.getParticle(i)){
                return (String)names.get(i);
            }
        }
        return "";
    }

    void setTarget(String particleName, float xTarget, float yTarget){
        int idx = names.indexOf(particleName);
        setTarget(idx, xTarget, yTarget);    
    }
    
    void setTarget(int idx, float xTarget, float yTarget){
        Smoother2D cent = (Smoother2D)centroids.get(idx);
        Particle p = physics.getParticle(idx);
        cent.setValue(p.position().x(), p.position().y());
        cent.setTarget(xTarget,yTarget);
    }
    
    
    void tick(){
        tick(1.0);
    }

    void tick(float dt){
        if (dt>0.0){
            anim.tick();
        }
        // For all the particles without -1,-1 targets, move them toward the target
        for (int i=0; i<physics.numberOfParticles(); i++){
            Smoother2D cent = (Smoother2D)centroids.get(i);
            Particle p = physics.getParticle(i);
            if (cent.getXTarget()==-1 && cent.getYTarget()==-1){
                p.makeFree();
            }else{
                p.makeFixed();
                p.moveTo(cent.x(),cent.y(),0);
            }
        }

        physics.tick(dt);
    }

    
    
}


void setup()
{
    size( 600, 400 );
    frameRate(30);
    frame = 0;
    textSystem = new SmartSystem( 0, .25, .93 );
    boxSizeSystem = new SmartSystem( 0, .25, .95 );
    boxSystem = new SmartSystem( 0.0, .25, .90 );
    floatSystem = new SmartSystem( 0.6, .25, .90 );
    htmlFont = loadFont("TheSansMono-SemiBold-12.vlw");
    renderFont = loadFont("Elzevir-Book-14.vlw");
    renderFontItal = loadFont("Elzevir-BookItalic-14.vlw");
    bgColor = color(#1c1b11);
        
    int lineX = 6;
    int lineY = 14;

    textFont(htmlFont,12);
    for (int i=0; i<htmlText.length; i++){
        textSystem.makeParticle("line"+i, lineX, lineY);
        lineX += textWidth(htmlText[i]);
        textSystem.setTarget(i,-1,-1);
    }
}

void arrangeHtml(){
    int baseline = 20;
    int typeX = 20;
    int typeY = 40;
    int typeIndent = 20;

    textSystem.setTarget(0,typeX,typeY);
    typeY += baseline;
    typeX += typeIndent;
    textSystem.setTarget(1,typeX,typeY);
    typeY += baseline;
    typeX += typeIndent;
    textSystem.setTarget(2,typeX,typeY);
    typeY += baseline;
    textSystem.setTarget(3,typeX,typeY);
    typeY += baseline;
    textSystem.setTarget(4,typeX,typeY);
    typeY += baseline;
    typeX -= typeIndent;
    textSystem.setTarget(5,typeX,typeY);
    typeY += baseline;
    typeX -= typeIndent;
    textSystem.setTarget(6,typeX,typeY);    
}

void growBoxes(){
    // make a particle system with with the locations representing
    // the growing width and height of the box for this element
    boxSystem.clear();
    boxSizeSystem.clear();
    
    // Set the rect origins
    boxSystem.makeParticle("html", 14,24);
    boxSystem.makeParticle("body", 34,46);
    boxSystem.makeParticle("p1", 54,69);
    boxSystem.makeParticle("p2", 54,89);
    boxSystem.makeParticle("p3", 54,109);
    boxSystem.makeParticle("em", 176,90);
    boxSystem.makeParticle("a", 109,110);

    // Set the target sizes appropriately
    boxSizeSystem.makeParticle("html", 0, 144);
    boxSizeSystem.setTarget(0,377,144); // html
    boxSizeSystem.makeParticle("body", 0, 100);
    boxSizeSystem.setTarget(1,350,100); // body
    boxSizeSystem.makeParticle("p1", 0, 15);
    boxSizeSystem.setTarget(2,227,15); // p1
    boxSizeSystem.makeParticle("p2", 0, 15);
    boxSizeSystem.setTarget(3,284,15); // p2
    boxSizeSystem.makeParticle("p3", 0, 15);
    boxSizeSystem.setTarget(4,324,15); // p3
    boxSizeSystem.makeParticle("em", 0, 13);
    boxSizeSystem.setTarget(5,130,13); // em
    boxSizeSystem.makeParticle("a", 0, 12);
    boxSizeSystem.setTarget(6,240,12); // a
}

void shrinkBoxes(){
    boxSystem.setTarget(0,16,28);
    boxSizeSystem.setTarget(0,47,15); // html
    boxSystem.setTarget(1,36,48);
    boxSizeSystem.setTarget(1,48,15); // body

    boxSystem.setTarget(2,58,70);
    boxSizeSystem.setTarget(2,25,14); // p1
    boxSystem.setTarget(3,58,90);
    boxSizeSystem.setTarget(3,25,14); // p2
    boxSystem.setTarget(4,58,110);
    boxSizeSystem.setTarget(4,25,14); // p3

    boxSystem.setTarget(5,178,90);
    boxSizeSystem.setTarget(5,30,13); // em
    boxSystem.setTarget(6,107,110);
    boxSizeSystem.setTarget(6,20,12); // a
}
    
void floatBoxes(){
    floatSystem.clear();

    for (int i=0; i<boxSystem.count(); i++){
        Particle p = boxSystem.getParticle(i);
        String tag = boxSystem.nameForParticle(p);
        
        floatSystem.makeParticle(tag, p.position().x()+10, p.position().y()+10);
        floatSystem.setTarget(i,-1,-1); // redundant? should it be?
    }

    floatSystem.linkNodes("html", "body");
    floatSystem.tick(0.1);
    floatSystem.linkNodes("body", "p1");
    Particle p1 = floatSystem.getParticle("p1");
    p1.addVelocity(0.0,-10,0.0);
    
    floatSystem.tick(0.01);
    floatSystem.linkNodes("body", "p2");
    floatSystem.tick(0.01);
    floatSystem.linkNodes("p2", "em");
    floatSystem.tick(0.01);
    floatSystem.linkNodes("body", "p3");
    floatSystem.tick(0.01);
    floatSystem.linkNodes("p3", "a");
    floatSystem.tick(0.01);
}

void mousePressed() {
    // reset to the drawing-boxes phase
    if (frame > 250){
        frame = 250;
    }
}

void draw()
{
    background(bgColor);
    textSystem.tick( 1.6 );
    boxSizeSystem.tick();
    boxSystem.tick();
    floatSystem.tick(0.2 );
    
    // Loop back to the box drawing once the tree floats around for a while
    if (frame>1800){ frame = 250; }
    
    // Handle the element box drawing
    if (frame==250){ growBoxes(); }
    if (frame==400){ shrinkBoxes(); }

    // Draw the floating tree
    if (frame==500){ floatBoxes(); }


    // Draw the element boxes
    if (frame>=250){
        noStroke();

        for (int i=0; i<7; i++){
            Particle p, s;
            p = boxSystem.getParticle(i);
            s = boxSizeSystem.getParticle(i);
            fill(255,255,255,30);
            if (i==0){
                fill(#38372e);
            }else if (i==1){
                fill(#52524a);
            }else if (i>1 && i<5){
                fill(#6b6b64); //60
            }else if (i==5){
                fill(#582d0b);
            } else if(i==6){
                fill(#122d51);
            }
            rect(p.position().x(),p.position().y(),
                 s.position().x(),s.position().y());
        }
    }





    // Draw the floating pieces of HTML text
    fill(255);
    textFont(htmlFont,12);
    if (frame<60){
        if(frame>20){
            fill(255,255,255,255*((frame-20)/40.0));
        }else{
            fill(255,255,255,0);            
        }
    }else if (frame==100){
        arrangeHtml();
    }

    // draw the ghost text floating over
    if (frame >170){
        color textColor = color(255,255,255,0);
        float offset = 350*(frame-170)/70.0;
        boolean shouldDraw = true;
        if (frame<220){
            textColor = color(255,255,255,20+50*(frame-170)/50.0);
        }else if (frame>=220 && frame < 235){
            textColor = color(255,255,255,20+50*(frame-230)/-15.0);
        }else{
            shouldDraw = false;
        }

        if(shouldDraw){
            float meanY = 0;
            float maxY = 0;
            float minY = height;
            for (int i=0; i<textSystem.count(); i++){
                Particle p = textSystem.getParticle(i);                
                meanY += p.position().y();
                maxY = max(maxY,p.position().y());
                minY = min(minY,p.position().y());
            }
            
            maxY *= 1.25;
            //minY *= 0.5;
            minY = 0;
                        
            meanY /= 1.0*textSystem.count();
            float animStep = (frame - 170)/70.0;
            animStep = min(animStep,1.0);
            for (int i=0; i<textSystem.count(); i++){
                Particle p = textSystem.getParticle(i);
                float targetY = maxY - (p.position().y() - minY);
                float easedY = animStep*targetY + (1.0-animStep)*p.position().y();
                
                fill(textColor);
                //println("from "+p.position().y()+" to "+easedY);
                text(htmlText[i],p.position().x()+offset + 30*(i%3), easedY);
            }
        }
    }


    // Draw the `rendered' HTML page
    if (frame > 200){
        color pageColor = color(255,255,255,0);
        color textColor = color(0,0,0,0);
        float alpha = (frame-200)/40.0;
        if(frame < 240){
            pageColor = color(235,235,235,200*alpha);
            textColor = color(40,40,40,255*alpha);
        }else{
            pageColor = color(235,235,235,200);
            textColor = color(40,40,40,255);
        }

        int gutter = 412;
        int hangline = 45;

        // blank out the page
        fill(pageColor);
        noStroke();
        rect(gutter-10,hangline-21,139,143);

        // draw the three lines of text
        fill(textColor);
        textFont(renderFont,14);
        text("this is just plain text,", gutter, hangline);

        text("but text with ", gutter, hangline+15);
        float lineWidth = textWidth("but text with ");
        textFont(renderFontItal,14);
        //fill(151,49,18,255*alpha);
        fill(88,45,11,255*alpha);
        text("structure", gutter+lineWidth, hangline+15);

        fill(textColor);
        textFont(renderFont,14);
        text("and ",gutter,hangline+30);
        lineWidth = textWidth("and ");
        fill(18,45,81,255*alpha);
        text("connections", gutter+lineWidth, hangline+30);
        float linkWidth = textWidth("connections");
        strokeWeight(1);
        stroke(18,45,81,100*alpha);
        line(gutter+lineWidth,hangline+31,gutter+lineWidth+linkWidth,hangline+31);
    }


    if (frame>=250 && floatSystem.count()>0){
        
        // kick things
        Particle body = floatSystem.getParticle("body");
        body.addVelocity(0.5,-0.25,0.0);

        Particle html = floatSystem.getParticle("html");
        html.setMass(2);
        html.addVelocity(0.01,-0,0.0);
        
        for (int i=0; i<floatSystem.count(); i++){
            Particle p = floatSystem.getParticle(i);
            if (p.position().x()>width*.4 && p.position().x()<width*.5){
                p.addVelocity(0,-0.08,0);
            }
        }      
        
          
        // draw edges
        stroke( 255,255,255, 30 );
        strokeWeight(2.0);
        beginShape( LINES );
        for ( int j = 0; j < floatSystem.sys().numberOfSprings(); ++j )
        {
          Spring e = floatSystem.sys().getSpring( j );
          Particle a = e.getOneEnd();
          Particle b = e.getTheOtherEnd();
          vertex( a.position().x(), a.position().y() );
          vertex( b.position().x(), b.position().y() );
        }
        endShape();

        // draw floating nodes
        ellipseMode(CENTER);
        for (int i=0; i<floatSystem.count(); i++){
            Particle p = floatSystem.getParticle(i);
            String pName = floatSystem.nameForParticle(p);
            noStroke();
            if (pName.equals("em")){
                fill(#582d0b);
            }else if(pName.equals("a")){
                fill(#122d51);
            }else if(pName.equals("p1") || pName.equals("p2") || pName.equals("p3")){
                    fill(#99998e);
            }else if(pName.equals("body")){
                fill(#52524a);
            }else{
                fill(#38372e);
            }
            ellipse(p.position().x(), p.position().y(),10+.08*p.age(),10+.08*p.age());
        }
    }

    // draw the html text
    noStroke();
    fill(255);
    textFont(htmlFont,12);
    for (int i=0; i<textSystem.count(); i++){
        Particle p = textSystem.getParticle(i);
        text(htmlText[i],p.position().x(), p.position().y());
    }
    textSystem.tick();



    frame++;
}
